@@ -1,23 +1,14 @@
-from abc import ABCMeta
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    ClassVar,
-    Generic,
-    Iterable,
-    Sequence,
-    TypeVar,
-    Union,
-)
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeVar
 
 from typing_extensions import final
 
-from returns._generated.iterable import iterable
 from returns.context import NoDeps
+from returns.interfaces.specific import reader_ioresult
 from returns.io import IO, IOFailure, IOResult, IOSuccess
 from returns.primitives.container import BaseContainer
-from returns.primitives.types import Immutable
+from returns.primitives.hkt import Kind3, SupportsKind3, dekind
 from returns.result import Result
 
 if TYPE_CHECKING:
@@ -26,6 +17,7 @@ if TYPE_CHECKING:
 
 # Context:
 _EnvType = TypeVar('_EnvType', contravariant=True)
+_NewEnvType = TypeVar('_NewEnvType')
 
 # Result:
 _ValueType = TypeVar('_ValueType', covariant=True)
@@ -40,7 +32,8 @@ _FirstType = TypeVar('_FirstType')
 @final
 class RequiresContextIOResult(
     BaseContainer,
-    Generic[_EnvType, _ValueType, _ErrorType],
+    SupportsKind3['RequiresContextIOResult', _ValueType, _ErrorType, _EnvType],
+    reader_ioresult.ReaderIOResultBasedN[_ValueType, _ErrorType, _EnvType],
 ):
     """
     The ``RequiresContextIOResult`` combinator.
@@ -50,7 +43,7 @@ class RequiresContextIOResult(
     for more docs.
 
     This is just a handy wrapper around
-    ``RequiresContext[env, IOResult[a, b]]``
+    ``RequiresContext[IOResult[a, b], env]``
     which represents a context-dependent impure operation that might fail.
 
     It has several important differences from the regular ``Result`` classes.
@@ -91,8 +84,9 @@ class RequiresContextIOResult(
 
     This is a complex type for complex tasks!
 
-    Important implementation detail:
-    due it is meaning, ``RequiresContextIOResult``
+    .. rubric:: Important implementation details
+
+    Due it is meaning, ``RequiresContextIOResult``
     cannot have ``Success`` and ``Failure`` subclasses.
 
     We only have just one type. That's by design.
@@ -101,10 +95,10 @@ class RequiresContextIOResult(
     Use converters inside the ``RequiresContext`` context, not outside.
 
     See also:
-        https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
-        https://en.wikipedia.org/wiki/Lazy_evaluation
-        https://bit.ly/2R8l4WK
-        https://bit.ly/2RwP4fp
+        - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
+        - https://en.wikipedia.org/wiki/Lazy_evaluation
+        - https://bit.ly/2R8l4WK
+        - https://bit.ly/2RwP4fp
 
     """
 
@@ -112,12 +106,12 @@ class RequiresContextIOResult(
     #: is just a function that returns `IOResult`.
     #: This field has an extra 'RequiresContext' just because `mypy` needs it.
     _inner_value: Callable[
-        ['RequiresContextIOResult', _EnvType],
+        [RequiresContextIOResult, _EnvType],
         IOResult[_ValueType, _ErrorType],
     ]
 
-    #: A convinient placeholder to call methods created by `.from_value()`.
-    empty: ClassVar[NoDeps] = object()
+    #: A convenient placeholder to call methods created by `.from_value()`.
+    no_args: ClassVar[NoDeps] = object()
 
     def __init__(
         self,
@@ -148,7 +142,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
           >>> from returns.io import IOSuccess
 
-          >>> def first(lg: bool) -> RequiresContextIOResult[float, int, str]:
+          >>> def first(lg: bool) -> RequiresContextIOResult[int, str, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextIOResult(
           ...         lambda deps: IOSuccess(deps if lg else -deps),
@@ -162,9 +156,34 @@ class RequiresContextIOResult(
         """
         return self._inner_value(deps)
 
+    def swap(
+        self,
+    ) -> RequiresContextIOResult[_ErrorType, _ValueType, _EnvType]:
+        """
+        Swaps value and error types.
+
+        So, values become errors and errors become values.
+        It is useful when you have to work with errors a lot.
+        And since we have a lot of ``.bind_`` related methods
+        and only a single ``.lash`` - it is easier to work with values.
+
+        .. code:: python
+
+          >>> from returns.context import RequiresContextIOResult
+          >>> from returns.io import IOSuccess, IOFailure
+
+          >>> success = RequiresContextIOResult.from_value(1)
+          >>> failure = RequiresContextIOResult.from_failure(1)
+
+          >>> assert success.swap()(...) == IOFailure(1)
+          >>> assert failure.swap()(...) == IOSuccess(1)
+
+        """
+        return RequiresContextIOResult(lambda deps: self(deps).swap())
+
     def map(  # noqa: WPS125
         self, function: Callable[[_ValueType], _NewValueType],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Composes successful container with a pure function.
 
@@ -186,16 +205,20 @@ class RequiresContextIOResult(
 
     def apply(
         self,
-        container: 'RequiresContextIOResult['
-            '_EnvType, Callable[[_ValueType], _NewValueType], _ErrorType]',
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+        container: Kind3[
+            RequiresContextIOResult,
+            Callable[[_ValueType], _NewValueType],
+            _ErrorType,
+            _EnvType,
+        ],
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Calls a wrapped function in a container on this container.
 
         .. code:: python
 
           >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IOSuccess, IOFailure, IOResult
+          >>> from returns.io import IOSuccess, IOFailure
 
           >>> def transform(arg: str) -> str:
           ...     return arg + 'b'
@@ -204,26 +227,35 @@ class RequiresContextIOResult(
           ...    RequiresContextIOResult.from_value(transform),
           ... )(...) == IOSuccess('ab')
 
+          >>> assert RequiresContextIOResult.from_value('a').apply(
+          ...    RequiresContextIOResult.from_failure(1),
+          ... )(...) == IOFailure(1)
+
           >>> assert RequiresContextIOResult.from_failure('a').apply(
           ...    RequiresContextIOResult.from_value(transform),
           ... )(...) == IOFailure('a')
 
-          >>> assert isinstance(RequiresContextIOResult.from_value('a').apply(
-          ...    RequiresContextIOResult.from_failure(transform),
-          ... )(...), IOResult.failure_type) is True
+          >>> assert RequiresContextIOResult.from_failure('a').apply(
+          ...    RequiresContextIOResult.from_failure('b'),
+          ... )(...) == IOFailure('a')
 
         """
         return RequiresContextIOResult(
-            lambda deps: self(deps).apply(container(deps)),
+            lambda deps: self(deps).apply(dekind(container)(deps)),
         )
 
     def bind(
         self,
         function: Callable[
             [_ValueType],
-            'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]',
+            Kind3[
+                RequiresContextIOResult,
+                _NewValueType,
+                _ErrorType,
+                _EnvType,
+            ],
         ],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Composes this container with a function returning the same type.
 
@@ -232,7 +264,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
           >>> from returns.io import IOSuccess, IOFailure
 
-          >>> def first(lg: bool) -> RequiresContextIOResult[float, int, int]:
+          >>> def first(lg: bool) -> RequiresContextIOResult[int, int, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextIOResult(
           ...         lambda deps: IOSuccess(deps) if lg else IOFailure(-deps),
@@ -240,7 +272,7 @@ class RequiresContextIOResult(
 
           >>> def second(
           ...     number: int,
-          ... ) -> RequiresContextIOResult[float, str, int]:
+          ... ) -> RequiresContextIOResult[str, int, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextIOResult(
           ...         lambda deps: IOSuccess('>=' if number >= deps else '<'),
@@ -252,14 +284,19 @@ class RequiresContextIOResult(
         """
         return RequiresContextIOResult(
             lambda deps: self(deps).bind(
-                lambda inner: function(inner)(deps),  # type: ignore[misc]
+                lambda inner: dekind(  # type: ignore[misc]
+                    function(inner),
+                )(deps),
             ),
         )
 
+    #: Alias for `bind_context_ioresult` method, it is the same as `bind` here.
+    bind_context_ioresult = bind
+
     def bind_result(
         self,
-        function: Callable[[_ValueType], 'Result[_NewValueType, _ErrorType]'],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+        function: Callable[[_ValueType], Result[_NewValueType, _ErrorType]],
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``Result`` returning function to the current container.
 
@@ -274,15 +311,15 @@ class RequiresContextIOResult(
 
           >>> assert RequiresContextIOResult.from_value(1).bind_result(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOSuccess(2)
+          ... )(RequiresContextIOResult.no_args) == IOSuccess(2)
 
           >>> assert RequiresContextIOResult.from_value(0).bind_result(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOFailure('<0')
+          ... )(RequiresContextIOResult.no_args) == IOFailure('<0')
 
           >>> assert RequiresContextIOResult.from_failure(':(').bind_result(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOFailure(':(')
+          ... )(RequiresContextIOResult.no_args) == IOFailure(':(')
 
         """
         return RequiresContextIOResult(
@@ -293,9 +330,9 @@ class RequiresContextIOResult(
         self,
         function: Callable[
             [_ValueType],
-            'RequiresContext[_EnvType, _NewValueType]',
+            'RequiresContext[_NewValueType, _EnvType]',
         ],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``RequiresContext`` returning function to current container.
 
@@ -304,7 +341,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContext
           >>> from returns.io import IOSuccess, IOFailure
 
-          >>> def function(arg: int) -> RequiresContext[str, int]:
+          >>> def function(arg: int) -> RequiresContext[int, str]:
           ...     return RequiresContext(lambda deps: len(deps) + arg)
 
           >>> assert function(2)('abc') == 5
@@ -328,9 +365,9 @@ class RequiresContextIOResult(
         self,
         function: Callable[
             [_ValueType],
-            'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]',
+            'RequiresContextResult[_NewValueType, _ErrorType, _EnvType]',
         ],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``RequiresContextResult`` returning function to the current one.
 
@@ -340,7 +377,7 @@ class RequiresContextIOResult(
           >>> from returns.io import IOSuccess, IOFailure
           >>> from returns.result import Success, Failure
 
-          >>> def function(arg: int) -> RequiresContextResult[str, int, int]:
+          >>> def function(arg: int) -> RequiresContextResult[int, int, str]:
           ...     if arg > 0:
           ...         return RequiresContextResult(
           ...             lambda deps: Success(len(deps) + arg),
@@ -380,7 +417,7 @@ class RequiresContextIOResult(
     def bind_io(
         self,
         function: Callable[[_ValueType], IO[_NewValueType]],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``IO`` returning function to the current container.
 
@@ -394,11 +431,11 @@ class RequiresContextIOResult(
 
           >>> assert RequiresContextIOResult.from_value(1).bind_io(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOSuccess('1')
+          ... )(RequiresContextIOResult.no_args) == IOSuccess('1')
 
           >>> assert RequiresContextIOResult.from_failure(1).bind_io(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOFailure(1)
+          ... )(RequiresContextIOResult.no_args) == IOFailure(1)
 
         """
         return RequiresContextIOResult(
@@ -408,7 +445,7 @@ class RequiresContextIOResult(
     def bind_ioresult(
         self,
         function: Callable[[_ValueType], IOResult[_NewValueType, _ErrorType]],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``IOResult`` returning function to the current container.
 
@@ -422,46 +459,24 @@ class RequiresContextIOResult(
 
           >>> assert RequiresContextIOResult.from_value(1).bind_ioresult(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOSuccess(2)
+          ... )(RequiresContextIOResult.no_args) == IOSuccess(2)
 
           >>> assert RequiresContextIOResult.from_value(0).bind_ioresult(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOFailure('<0')
+          ... )(RequiresContextIOResult.no_args) == IOFailure('<0')
 
           >>> assert RequiresContextIOResult.from_failure(':(').bind_ioresult(
           ...     function,
-          ... )(RequiresContextIOResult.empty) == IOFailure(':(')
+          ... )(RequiresContextIOResult.no_args) == IOFailure(':(')
 
         """
         return RequiresContextIOResult(
             lambda deps: self(deps).bind(function),
         )
 
-    def fix(
-        self, function: Callable[[_ErrorType], _NewValueType],
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _ErrorType]':
-        """
-        Composes failed container with a pure function.
-
-        .. code:: python
-
-          >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IOSuccess
-
-          >>> assert RequiresContextIOResult.from_value(1).fix(
-          ...     lambda x: x + 1,
-          ... )(...) == IOSuccess(1)
-
-          >>> assert RequiresContextIOResult.from_failure(1).fix(
-          ...     lambda x: x + 1,
-          ... )(...) == IOSuccess(2)
-
-        """
-        return RequiresContextIOResult(lambda deps: self(deps).fix(function))
-
     def alt(
         self, function: Callable[[_ErrorType], _NewErrorType],
-    ) -> 'RequiresContextIOResult[_EnvType, _ValueType, _NewErrorType]':
+    ) -> RequiresContextIOResult[_ValueType, _NewErrorType, _EnvType]:
         """
         Composes failed container with a pure function.
 
@@ -481,13 +496,18 @@ class RequiresContextIOResult(
         """
         return RequiresContextIOResult(lambda deps: self(deps).alt(function))
 
-    def rescue(
+    def lash(
         self,
         function: Callable[
             [_ErrorType],
-            'RequiresContextIOResult[_EnvType, _ValueType, _NewErrorType]',
+            Kind3[
+                RequiresContextIOResult,
+                _ValueType,
+                _NewErrorType,
+                _EnvType,
+            ],
         ],
-    ) -> 'RequiresContextIOResult[_EnvType, _ValueType, _NewErrorType]':
+    ) -> RequiresContextIOResult[_ValueType, _NewErrorType, _EnvType]:
         """
         Composes this container with a function returning the same type.
 
@@ -496,7 +516,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
           >>> from returns.io import IOSuccess, IOFailure
 
-          >>> def rescuable(
+          >>> def lashable(
           ...     arg: str,
           ... ) -> RequiresContextIOResult[str, str, str]:
           ...      if len(arg) > 1:
@@ -507,95 +527,117 @@ class RequiresContextIOResult(
           ...          lambda deps: IOFailure(arg + deps),
           ...      )
 
-          >>> assert RequiresContextIOResult.from_value('a').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextIOResult.from_value('a').lash(
+          ...     lashable,
           ... )('c') == IOSuccess('a')
-          >>> assert RequiresContextIOResult.from_failure('a').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextIOResult.from_failure('a').lash(
+          ...     lashable,
           ... )('c') == IOFailure('ac')
-          >>> assert RequiresContextIOResult.from_failure('aa').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextIOResult.from_failure('aa').lash(
+          ...     lashable,
           ... )('b') == IOSuccess('baa')
 
         """
         return RequiresContextIOResult(
-            lambda deps: self(deps).rescue(
-                lambda inner: function(inner)(deps),  # type: ignore[misc]
+            lambda deps: self(deps).lash(
+                lambda inner: function(inner)(deps),  # type: ignore
             ),
         )
 
-    def value_or(
-        self, default_value: _FirstType,
-    ) -> Callable[[_EnvType], IO[Union[_ValueType, _FirstType]]]:
+    def compose_result(
+        self,
+        function: Callable[
+            [Result[_ValueType, _ErrorType]],
+            Kind3[
+                RequiresContextIOResult,
+                _NewValueType,
+                _ErrorType,
+                _EnvType,
+            ],
+        ],
+    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType]:
         """
-        Returns a callable that either returns a success or default value.
+        Composes inner ``Result`` with ``ReaderIOResult`` returning function.
+
+        Can be useful when you need an access to both states of the result.
 
         .. code:: python
 
-          >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IO
+          >>> from returns.context import ReaderIOResult, NoDeps
+          >>> from returns.io import IOSuccess, IOFailure
+          >>> from returns.result import Result
 
-          >>> assert RequiresContextIOResult.from_value(1).value_or(2)(
-          ...     RequiresContextIOResult.empty,
-          ... ) == IO(1)
+          >>> def count(
+          ...    container: Result[int, int],
+          ... ) -> ReaderIOResult[int, int, NoDeps]:
+          ...     return ReaderIOResult.from_result(
+          ...         container.map(lambda x: x + 1).alt(abs),
+          ...     )
 
-          >>> assert RequiresContextIOResult.from_failure(1).value_or(2)(
-          ...     RequiresContextIOResult.empty,
-          ... ) == IO(2)
+          >>> success = ReaderIOResult.from_value(1)
+          >>> failure = ReaderIOResult.from_failure(-1)
+          >>> assert success.compose_result(count)(...) == IOSuccess(2)
+          >>> assert failure.compose_result(count)(...) == IOFailure(1)
 
         """
-        return lambda deps: self(deps).value_or(default_value)
+        return RequiresContextIOResult(
+            lambda deps: dekind(
+                function(self(deps)._inner_value),  # noqa: WPS437
+            )(deps),
+        )
 
-    def unwrap(self) -> Callable[[_EnvType], IO[_ValueType]]:
+    def modify_env(
+        self,
+        function: Callable[[_NewEnvType], _EnvType],
+    ) -> RequiresContextIOResult[_ValueType, _ErrorType, _NewEnvType]:
         """
-        Returns a callable that unwraps success value or raises exception.
+        Allows to modify the environment type.
 
-        .. code:: pycon
+        .. code:: python
 
-          >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IO
+          >>> from returns.context import RequiresContextIOResultE
+          >>> from returns.io import IOSuccess, impure_safe
 
-          >>> assert RequiresContextIOResult.from_value(1).unwrap()(
-          ...     RequiresContextIOResult.empty,
-          ... ) == IO(1)
+          >>> def div(arg: int) -> RequiresContextIOResultE[float, int]:
+          ...     return RequiresContextIOResultE(
+          ...         impure_safe(lambda deps: arg / deps),
+          ...     )
 
-          >>> RequiresContextIOResult.from_failure(1).unwrap()(
-          ...     RequiresContextIOResult.empty,
-          ... )
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
+          >>> assert div(3).modify_env(int)('2') == IOSuccess(1.5)
+          >>> assert div(3).modify_env(int)('0').failure()
 
         """
-        return lambda deps: self(deps).unwrap()
+        return RequiresContextIOResult(lambda deps: self(function(deps)))
 
-    def failure(self) -> Callable[[_EnvType], IO[_ErrorType]]:
+    @classmethod
+    def ask(cls) -> RequiresContextIOResult[_EnvType, _ErrorType, _EnvType]:
         """
-        Returns a callable that unwraps failure value or raises exception.
+        Is used to get the current dependencies inside the call stack.
 
-        .. code:: pycon
+        Similar to :meth:`returns.context.requires_context.RequiresContext.ask`,
+        but returns ``IOResult`` instead of a regular value.
 
-          >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IO
+        Please, refer to the docs there to learn how to use it.
 
-          >>> assert RequiresContextIOResult.from_failure(1).failure()(
-          ...     RequiresContextIOResult.empty,
-          ... ) == IO(1)
+        One important note that is worth duplicating here:
+        you might need to provide ``_EnvType`` explicitly,
+        so ``mypy`` will know about it statically.
 
-          >>> RequiresContextIOResult.from_value(1).failure()(
-          ...    RequiresContextIOResult.empty,
-          ... )
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
+        .. code:: python
+
+          >>> from returns.context import RequiresContextIOResultE
+          >>> from returns.io import IOSuccess
+          >>> assert RequiresContextIOResultE[int, int].ask().map(
+          ...     str,
+          ... )(1) == IOSuccess('1')
 
         """
-        return lambda deps: self(deps).failure()
+        return RequiresContextIOResult(IOSuccess)
 
     @classmethod
     def from_result(
-        cls, inner_value: 'Result[_ValueType, _ErrorType]',
-    ) -> 'RequiresContextIOResult[NoDeps, _ValueType, _ErrorType]':
+        cls, inner_value: Result[_NewValueType, _NewErrorType],
+    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, NoDeps]:
         """
         Creates new container with ``Result`` as a unit value.
 
@@ -604,7 +646,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
           >>> from returns.result import Success, Failure
           >>> from returns.io import IOSuccess, IOFailure
-          >>> deps = RequiresContextIOResult.empty
+          >>> deps = RequiresContextIOResult.no_args
 
           >>> assert RequiresContextIOResult.from_result(
           ...    Success(1),
@@ -623,7 +665,7 @@ class RequiresContextIOResult(
     def from_io(
         cls,
         inner_value: IO[_NewValueType],
-    ) -> 'RequiresContextIOResult[NoDeps, _NewValueType, Any]':
+    ) -> RequiresContextIOResult[_NewValueType, Any, NoDeps]:
         """
         Creates new container from successful ``IO`` value.
 
@@ -633,7 +675,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
 
           >>> assert RequiresContextIOResult.from_io(IO(1))(
-          ...     RequiresContextIOResult.empty,
+          ...     RequiresContextIOResult.no_args,
           ... ) == IOSuccess(1)
 
         """
@@ -645,7 +687,7 @@ class RequiresContextIOResult(
     def from_failed_io(
         cls,
         inner_value: IO[_NewErrorType],
-    ) -> 'RequiresContextIOResult[NoDeps, Any, _NewErrorType]':
+    ) -> RequiresContextIOResult[Any, _NewErrorType, NoDeps]:
         """
         Creates a new container from failed ``IO`` value.
 
@@ -655,7 +697,7 @@ class RequiresContextIOResult(
           >>> from returns.context import RequiresContextIOResult
 
           >>> assert RequiresContextIOResult.from_failed_io(IO(1))(
-          ...     RequiresContextIOResult.empty,
+          ...     RequiresContextIOResult.no_args,
           ... ) == IOFailure(1)
 
         """
@@ -665,8 +707,8 @@ class RequiresContextIOResult(
 
     @classmethod
     def from_ioresult(
-        cls, inner_value: IOResult[_ValueType, _ErrorType],
-    ) -> 'RequiresContextIOResult[NoDeps, _ValueType, _ErrorType]':
+        cls, inner_value: IOResult[_NewValueType, _NewErrorType],
+    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, NoDeps]:
         """
         Creates new container with ``IOResult`` as a unit value.
 
@@ -674,7 +716,7 @@ class RequiresContextIOResult(
 
           >>> from returns.context import RequiresContextIOResult
           >>> from returns.io import IOSuccess, IOFailure
-          >>> deps = RequiresContextIOResult.empty
+          >>> deps = RequiresContextIOResult.no_args
 
           >>> assert RequiresContextIOResult.from_ioresult(
           ...    IOSuccess(1),
@@ -687,12 +729,37 @@ class RequiresContextIOResult(
         """
         return RequiresContextIOResult(lambda _: inner_value)
 
+    @classmethod
+    def from_ioresult_context(
+        cls,
+        inner_value:
+            ReaderIOResult[_NewValueType, _NewErrorType, _NewEnvType],
+    ) -> ReaderIOResult[_NewValueType, _NewErrorType, _NewEnvType]:
+        """
+        Creates new container with ``ReaderIOResult`` as a unit value.
+
+        .. code:: python
+
+          >>> from returns.context import RequiresContextIOResult
+          >>> from returns.io import IOSuccess, IOFailure
+
+          >>> assert RequiresContextIOResult.from_ioresult_context(
+          ...    RequiresContextIOResult.from_value(1),
+          ... )(...) == IOSuccess(1)
+
+          >>> assert RequiresContextIOResult.from_ioresult_context(
+          ...    RequiresContextIOResult.from_failure(1),
+          ... )(...) == IOFailure(1)
+
+        """
+        return inner_value
+
     @classmethod
     def from_typecast(
         cls,
         inner_value:
-            'RequiresContext[_EnvType, IOResult[_NewValueType, _NewErrorType]]',
-    ) -> 'RequiresContextIOResult[_EnvType, _NewValueType, _NewErrorType]':
+            'RequiresContext[IOResult[_NewValueType, _NewErrorType], _EnvType]',
+    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, _EnvType]:
         """
         You might end up with ``RequiresContext[IOResult]`` as a value.
 
@@ -708,19 +775,19 @@ class RequiresContextIOResult(
 
           >>> assert RequiresContextIOResult.from_typecast(
           ...     RequiresContext.from_value(IOSuccess(1)),
-          ... )(RequiresContextIOResult.empty) == IOSuccess(1)
+          ... )(RequiresContextIOResult.no_args) == IOSuccess(1)
 
           >>> assert RequiresContextIOResult.from_typecast(
           ...     RequiresContext.from_value(IOFailure(1)),
-          ... )(RequiresContextIOResult.empty) == IOFailure(1)
+          ... )(RequiresContextIOResult.no_args) == IOFailure(1)
 
         """
         return RequiresContextIOResult(inner_value)
 
     @classmethod
     def from_context(
-        cls, inner_value: 'RequiresContext[_EnvType, _FirstType]',
-    ) -> 'RequiresContextIOResult[_EnvType, _FirstType, Any]':
+        cls, inner_value: 'RequiresContext[_NewValueType, _NewEnvType]',
+    ) -> RequiresContextIOResult[_NewValueType, Any, _NewEnvType]:
         """
         Creates new container from ``RequiresContext`` as a success unit.
 
@@ -740,8 +807,8 @@ class RequiresContextIOResult(
 
     @classmethod
     def from_failed_context(
-        cls, inner_value: 'RequiresContext[_EnvType, _FirstType]',
-    ) -> 'RequiresContextIOResult[_EnvType, Any, _FirstType]':
+        cls, inner_value: 'RequiresContext[_NewValueType, _NewEnvType]',
+    ) -> RequiresContextIOResult[Any, _NewValueType, _NewEnvType]:
         """
         Creates new container from ``RequiresContext`` as a failure unit.
 
@@ -762,8 +829,9 @@ class RequiresContextIOResult(
     @classmethod
     def from_result_context(
         cls,
-        inner_value: 'RequiresContextResult[_EnvType, _ValueType, _ErrorType]',
-    ) -> 'RequiresContextIOResult[_EnvType, _ValueType, _ErrorType]':
+        inner_value:
+            'RequiresContextResult[_NewValueType, _NewErrorType, _NewEnvType]',
+    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, _NewEnvType]:
         """
         Creates new container from ``RequiresContextResult`` as a unit value.
 
@@ -787,8 +855,9 @@ class RequiresContextIOResult(
 
     @classmethod
     def from_value(
-        cls, inner_value: _FirstType,
-    ) -> 'RequiresContextIOResult[NoDeps, _FirstType, Any]':
+        cls,
+        inner_value: _NewValueType,
+    ) -> RequiresContextIOResult[_NewValueType, Any, NoDeps]:
         """
         Creates new container with ``IOSuccess(inner_value)`` as a unit value.
 
@@ -798,7 +867,7 @@ class RequiresContextIOResult(
           >>> from returns.io import IOSuccess
 
           >>> assert RequiresContextIOResult.from_value(1)(
-          ...    RequiresContextIOResult.empty,
+          ...    RequiresContextIOResult.no_args,
           ... ) == IOSuccess(1)
 
         """
@@ -806,8 +875,9 @@ class RequiresContextIOResult(
 
     @classmethod
     def from_failure(
-        cls, inner_value: _FirstType,
-    ) -> 'RequiresContextIOResult[NoDeps, Any, _FirstType]':
+        cls,
+        inner_value: _NewErrorType,
+    ) -> RequiresContextIOResult[Any, _NewErrorType, NoDeps]:
         """
         Creates new container with ``IOFailure(inner_value)`` as a unit value.
 
@@ -817,94 +887,22 @@ class RequiresContextIOResult(
           >>> from returns.io import IOFailure
 
           >>> assert RequiresContextIOResult.from_failure(1)(
-          ...     RequiresContextIOResult.empty,
+          ...     RequiresContextIOResult.no_args,
           ... ) == IOFailure(1)
 
         """
         return RequiresContextIOResult(lambda _: IOFailure(inner_value))
 
-    @classmethod
-    def from_iterable(
-        cls,
-        inner_value:
-            Iterable[
-                'RequiresContextIOResult[_EnvType, _ValueType, _ErrorType]',
-            ],
-    ) -> 'RequiresContextIOResult[_EnvType, Sequence[_ValueType], _ErrorType]':
-        """
-        Transforms an iterable of ``RequiresContextIOResult`` containers.
-
-        Returns a single container with multiple elements inside.
-
-        .. code:: python
-
-          >>> from returns.context import RequiresContextIOResult
-          >>> from returns.io import IOSuccess, IOFailure
-
-          >>> assert RequiresContextIOResult.from_iterable([
-          ...    RequiresContextIOResult.from_value(1),
-          ...    RequiresContextIOResult.from_value(2),
-          ... ])(...) == IOSuccess((1, 2))
-
-          >>> assert RequiresContextIOResult.from_iterable([
-          ...    RequiresContextIOResult.from_value(1),
-          ...    RequiresContextIOResult.from_failure('a'),
-          ... ])(...) == IOFailure('a')
-
-          >>> assert RequiresContextIOResult.from_iterable([
-          ...    RequiresContextIOResult.from_failure('a'),
-          ...    RequiresContextIOResult.from_value(1),
-          ... ])(...) == IOFailure('a')
-
-        """
-        return iterable(cls, inner_value)
-
-
-@final
-class ContextIOResult(Immutable, Generic[_EnvType], metaclass=ABCMeta):
-    """
-    Helpers that can be used to work with ``RequiresContextIOResult`` container.
-
-    Related to :class:`returns.context.requires_context.Context`
-    and :class:`returns.context.requires_context_result.ContextResult`,
-    refer there for the docs.
-    """
-
-    __slots__ = ()
-
-    @classmethod
-    def ask(cls) -> RequiresContextIOResult[_EnvType, _EnvType, Any]:
-        """
-        Is used to get the current dependencies inside the call stack.
-
-        Similar to :meth:`returns.context.requires_context.Context.ask`,
-        but returns ``IOResult`` instead of a regular value.
-
-        Please, refer to the docs there to learn how to use it.
-
-        One important note that is worth duplicating here:
-        you might need to provide ``_EnvType`` explicitly,
-        so ``mypy`` will know about it statically.
-
-        .. code:: python
-
-          >>> from returns.context import ContextIOResult
-          >>> from returns.io import IOSuccess
-          >>> assert ContextIOResult[int].ask().map(str)(1) == IOSuccess('1')
-
-        """
-        return RequiresContextIOResult(IOSuccess)
-
 
 # Aliases:
 
 #: Alias for a popular case when ``Result`` has ``Exception`` as error type.
 RequiresContextIOResultE = RequiresContextIOResult[
-    _EnvType, _ValueType, Exception,
+    _ValueType, Exception, _EnvType,
 ]
 
 #: Alias to save you some typing. Uses original name from Haskell.
-ReaderIOResult = RequiresContextIOResult[_EnvType, _ValueType, _ErrorType]
+ReaderIOResult = RequiresContextIOResult
 
 #: Alias to save you some typing. Uses ``Exception`` as error type.
-ReaderIOResultE = RequiresContextIOResult[_EnvType, _ValueType, Exception]
+ReaderIOResultE = RequiresContextIOResult[_ValueType, Exception, _EnvType]
