@@ -1,22 +1,13 @@
-from abc import ABCMeta
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    ClassVar,
-    Generic,
-    Iterable,
-    Sequence,
-    TypeVar,
-    Union,
-)
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeVar
 
 from typing_extensions import final
 
-from returns._generated.iterable import iterable
 from returns.context import NoDeps
+from returns.interfaces.specific import reader_result
 from returns.primitives.container import BaseContainer
-from returns.primitives.types import Immutable
+from returns.primitives.hkt import Kind3, SupportsKind3, dekind
 from returns.result import Failure, Result, Success
 
 if TYPE_CHECKING:
@@ -24,6 +15,7 @@ if TYPE_CHECKING:
 
 # Context:
 _EnvType = TypeVar('_EnvType', contravariant=True)
+_NewEnvType = TypeVar('_NewEnvType')
 
 # Result:
 _ValueType = TypeVar('_ValueType', covariant=True)
@@ -38,14 +30,15 @@ _FirstType = TypeVar('_FirstType')
 @final
 class RequiresContextResult(
     BaseContainer,
-    Generic[_EnvType, _ValueType, _ErrorType],
+    SupportsKind3['RequiresContextResult', _ValueType, _ErrorType, _EnvType],
+    reader_result.ReaderResultBasedN[_ValueType, _ErrorType, _EnvType],
 ):
     """
     The ``RequiresContextResult`` combinator.
 
     See :class:`returns.context.requires_context.RequiresContext` for more docs.
 
-    This is just a handy wrapper around ``RequiresContext[env, Result[a, b]]``
+    This is just a handy wrapper around ``RequiresContext[Result[a, b], env]``
     which represents a context-dependent pure operation
     that might fail and return :class:`returns.result.Result`.
 
@@ -82,8 +75,9 @@ class RequiresContextResult(
     - ``RequiresContext`` values and pure functions returning it
     - ``Result`` and functions returning it
 
-    Important implementation detail:
-    due it is meaning, ``RequiresContextResult``
+    .. rubric:: Important implementation details
+
+    Due it is meaning, ``RequiresContextResult``
     cannot have ``Success`` and ``Failure`` subclasses.
 
     We only have just one type. That's by design.
@@ -92,21 +86,21 @@ class RequiresContextResult(
     Use converters inside the ``RequiresContext`` context, not outside.
 
     See also:
-        https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
-        https://en.wikipedia.org/wiki/Lazy_evaluation
-        https://bit.ly/2R8l4WK
-        https://bit.ly/2RwP4fp
+        - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
+        - https://en.wikipedia.org/wiki/Lazy_evaluation
+        - https://bit.ly/2R8l4WK
+        - https://bit.ly/2RwP4fp
 
     """
 
     #: This field has an extra 'RequiresContext' just because `mypy` needs it.
     _inner_value: Callable[
-        ['RequiresContextResult', _EnvType],
+        [RequiresContextResult, _EnvType],
         Result[_ValueType, _ErrorType],
     ]
 
-    #: A convinient placeholder to call methods created by `.from_value()`.
-    empty: ClassVar[NoDeps] = object()
+    #: A convenient placeholder to call methods created by `.from_value()`.
+    no_args: ClassVar[NoDeps] = object()
 
     def __init__(
         self,
@@ -137,7 +131,7 @@ class RequiresContextResult(
           >>> from returns.context import RequiresContextResult
           >>> from returns.result import Success
 
-          >>> def first(lg: bool) -> RequiresContextResult[float, int, str]:
+          >>> def first(lg: bool) -> RequiresContextResult[int, str, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextResult(
           ...         lambda deps: Success(deps if lg else -deps),
@@ -151,9 +145,32 @@ class RequiresContextResult(
         """
         return self._inner_value(deps)
 
+    def swap(self) -> RequiresContextResult[_ErrorType, _ValueType, _EnvType]:
+        """
+        Swaps value and error types.
+
+        So, values become errors and errors become values.
+        It is useful when you have to work with errors a lot.
+        And since we have a lot of ``.bind_`` related methods
+        and only a single ``.lash`` - it is easier to work with values.
+
+        .. code:: python
+
+          >>> from returns.context import RequiresContextResult
+          >>> from returns.result import Failure, Success
+
+          >>> success = RequiresContextResult.from_value(1)
+          >>> failure = RequiresContextResult.from_failure(1)
+
+          >>> assert success.swap()(...) == Failure(1)
+          >>> assert failure.swap()(...) == Success(1)
+
+        """
+        return RequiresContextResult(lambda deps: self(deps).swap())
+
     def map(  # noqa: WPS125
         self, function: Callable[[_ValueType], _NewValueType],
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Composes successful container with a pure function.
 
@@ -175,9 +192,13 @@ class RequiresContextResult(
 
     def apply(
         self,
-        container: 'RequiresContextResult['
-            '_EnvType, Callable[[_ValueType], _NewValueType], _ErrorType]',
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
+        container: Kind3[
+            RequiresContextResult,
+            Callable[[_ValueType], _NewValueType],
+            _ErrorType,
+            _EnvType,
+        ],
+    ) -> RequiresContextResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Calls a wrapped function in a container on this container.
 
@@ -203,16 +224,21 @@ class RequiresContextResult(
 
         """
         return RequiresContextResult(
-            lambda deps: self(deps).apply(container(deps)),
+            lambda deps: self(deps).apply(dekind(container)(deps)),
         )
 
     def bind(
         self,
         function: Callable[
             [_ValueType],
-            'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]',
+            Kind3[
+                RequiresContextResult,
+                _NewValueType,
+                _ErrorType,
+                _EnvType,
+            ],
         ],
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Composes this container with a function returning the same type.
 
@@ -221,7 +247,7 @@ class RequiresContextResult(
           >>> from returns.context import RequiresContextResult
           >>> from returns.result import Success, Failure
 
-          >>> def first(lg: bool) -> RequiresContextResult[float, int, int]:
+          >>> def first(lg: bool) -> RequiresContextResult[int, int, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextResult(
           ...         lambda deps: Success(deps) if lg else Failure(-deps),
@@ -229,7 +255,7 @@ class RequiresContextResult(
 
           >>> def second(
           ...     number: int,
-          ... ) -> RequiresContextResult[float, str, int]:
+          ... ) -> RequiresContextResult[str, int, float]:
           ...     # `deps` has `float` type here:
           ...     return RequiresContextResult(
           ...         lambda deps: Success('>=' if number >= deps else '<'),
@@ -245,10 +271,13 @@ class RequiresContextResult(
             ),
         )
 
+    #: Alias for `bind_context_result` method, it is the same as `bind` here.
+    bind_context_result = bind
+
     def bind_result(
         self,
         function: Callable[[_ValueType], Result[_NewValueType, _ErrorType]],
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``Result`` returning function to current container.
 
@@ -257,20 +286,20 @@ class RequiresContextResult(
           >>> from returns.context import RequiresContextResult
           >>> from returns.result import Success, Failure, Result
 
-          >>> def function(num: int) -> Result[int, str]:
+          >>> def function(num: int) -> Result[str, int]:
           ...     return Success(num + 1) if num > 0 else Failure('<0')
 
           >>> assert RequiresContextResult.from_value(1).bind_result(
           ...     function,
-          ... )(RequiresContextResult.empty) == Success(2)
+          ... )(RequiresContextResult.no_args) == Success(2)
 
           >>> assert RequiresContextResult.from_value(0).bind_result(
           ...     function,
-          ... )(RequiresContextResult.empty) == Failure('<0')
+          ... )(RequiresContextResult.no_args) == Failure('<0')
 
           >>> assert RequiresContextResult.from_failure(':(').bind_result(
           ...     function,
-          ... )(RequiresContextResult.empty) == Failure(':(')
+          ... )(RequiresContextResult.no_args) == Failure(':(')
 
         """
         return RequiresContextResult(lambda deps: self(deps).bind(function))
@@ -279,9 +308,9 @@ class RequiresContextResult(
         self,
         function: Callable[
             [_ValueType],
-            'RequiresContext[_EnvType, _NewValueType]',
+            'RequiresContext[_NewValueType, _EnvType]',
         ],
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
+    ) -> RequiresContextResult[_NewValueType, _ErrorType, _EnvType]:
         """
         Binds ``RequiresContext`` returning function to current container.
 
@@ -290,7 +319,7 @@ class RequiresContextResult(
           >>> from returns.context import RequiresContext
           >>> from returns.result import Success, Failure
 
-          >>> def function(arg: int) -> RequiresContext[str, int]:
+          >>> def function(arg: int) -> RequiresContext[int, str]:
           ...     return RequiresContext(lambda deps: len(deps) + arg)
 
           >>> assert function(2)('abc') == 5
@@ -310,31 +339,9 @@ class RequiresContextResult(
             ),
         )
 
-    def fix(
-        self, function: Callable[[_ErrorType], _NewValueType],
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _ErrorType]':
-        """
-        Composes failed container with a pure function.
-
-        .. code:: python
-
-          >>> from returns.context import RequiresContextResult
-          >>> from returns.result import Success
-
-          >>> assert RequiresContextResult.from_value(1).fix(
-          ...     lambda x: x + 1,
-          ... )(...) == Success(1)
-
-          >>> assert RequiresContextResult.from_failure(1).fix(
-          ...     lambda x: x + 1,
-          ... )(...) == Success(2)
-
-        """
-        return RequiresContextResult(lambda deps: self(deps).fix(function))
-
     def alt(
         self, function: Callable[[_ErrorType], _NewErrorType],
-    ) -> 'RequiresContextResult[_EnvType, _ValueType, _NewErrorType]':
+    ) -> RequiresContextResult[_ValueType, _NewErrorType, _EnvType]:
         """
         Composes failed container with a pure function.
 
@@ -354,13 +361,18 @@ class RequiresContextResult(
         """
         return RequiresContextResult(lambda deps: self(deps).alt(function))
 
-    def rescue(
+    def lash(
         self,
         function: Callable[
             [_ErrorType],
-            'RequiresContextResult[_EnvType, _ValueType, _NewErrorType]',
+            Kind3[
+                RequiresContextResult,
+                _ValueType,
+                _NewErrorType,
+                _EnvType,
+            ],
         ],
-    ) -> 'RequiresContextResult[_EnvType, _ValueType, _NewErrorType]':
+    ) -> RequiresContextResult[_ValueType, _NewErrorType, _EnvType]:
         """
         Composes this container with a function returning the same type.
 
@@ -369,7 +381,7 @@ class RequiresContextResult(
           >>> from returns.context import RequiresContextResult
           >>> from returns.result import Success, Failure
 
-          >>> def rescuable(arg: str) -> RequiresContextResult[str, str, str]:
+          >>> def lashable(arg: str) -> RequiresContextResult[str, str, str]:
           ...      if len(arg) > 1:
           ...          return RequiresContextResult(
           ...              lambda deps: Success(deps + arg),
@@ -378,94 +390,75 @@ class RequiresContextResult(
           ...          lambda deps: Failure(arg + deps),
           ...      )
 
-          >>> assert RequiresContextResult.from_value('a').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextResult.from_value('a').lash(
+          ...     lashable,
           ... )('c') == Success('a')
-          >>> assert RequiresContextResult.from_failure('a').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextResult.from_failure('a').lash(
+          ...     lashable,
           ... )('c') == Failure('ac')
-          >>> assert RequiresContextResult.from_failure('aa').rescue(
-          ...     rescuable,
+          >>> assert RequiresContextResult.from_failure('aa').lash(
+          ...     lashable,
           ... )('b') == Success('baa')
 
         """
         return RequiresContextResult(
-            lambda deps: self(deps).rescue(
+            lambda deps: self(deps).lash(
                 lambda inner: function(inner)(deps),  # type: ignore
             ),
         )
 
-    def value_or(
-        self, default_value: _FirstType,
-    ) -> Callable[[_EnvType], Union[_ValueType, _FirstType]]:
+    def modify_env(
+        self,
+        function: Callable[[_NewEnvType], _EnvType],
+    ) -> RequiresContextResult[_ValueType, _ErrorType, _NewEnvType]:
         """
-        Returns a callable that either returns a success or default value.
+        Allows to modify the environment type.
 
         .. code:: python
 
-          >>> from returns.context import RequiresContextResult
+          >>> from returns.context import RequiresContextResultE
+          >>> from returns.result import Success, safe
 
-          >>> assert RequiresContextResult.from_value(1).value_or(2)(
-          ...     RequiresContextResult.empty,
-          ... ) == 1
+          >>> def div(arg: int) -> RequiresContextResultE[float, int]:
+          ...     return RequiresContextResultE(
+          ...         safe(lambda deps: arg / deps),
+          ...     )
 
-          >>> assert RequiresContextResult.from_failure(1).value_or(2)(
-          ...     RequiresContextResult.empty,
-          ... ) == 2
+          >>> assert div(3).modify_env(int)('2') == Success(1.5)
+          >>> assert div(3).modify_env(int)('0').failure()
 
         """
-        return lambda deps: self(deps).value_or(default_value)
+        return RequiresContextResult(lambda deps: self(function(deps)))
 
-    def unwrap(self) -> Callable[[_EnvType], _ValueType]:
+    @classmethod
+    def ask(cls) -> RequiresContextResult[_EnvType, _ErrorType, _EnvType]:
         """
-        Returns a callable that unwraps success value or raises exception.
+        Is used to get the current dependencies inside the call stack.
 
-        .. code:: pycon
+        Similar to :meth:`returns.context.requires_context.RequiresContext.ask`,
+        but returns ``Result`` instead of a regular value.
 
-          >>> from returns.context import RequiresContextResult
-          >>> from returns.result import Success, Failure
+        Please, refer to the docs there to learn how to use it.
 
-          >>> assert RequiresContextResult(
-          ...    lambda _: Success(1),
-          ... ).unwrap()(RequiresContextResult.empty) == 1
+        One important note that is worth duplicating here:
+        you might need to provide ``_EnvType`` explicitly,
+        so ``mypy`` will know about it statically.
 
-          >>> RequiresContextResult(
-          ...    lambda _: Failure(1),
-          ... ).unwrap()(RequiresContextResult.empty)
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
+        .. code:: python
+
+          >>> from returns.context import RequiresContextResultE
+          >>> from returns.result import Success
+          >>> assert RequiresContextResultE[int, int].ask().map(
+          ...    str,
+          ... )(1) == Success('1')
 
         """
-        return lambda deps: self(deps).unwrap()
-
-    def failure(self) -> Callable[[_EnvType], _ErrorType]:
-        """
-        Returns a callable that unwraps failure value or raises exception.
-
-        .. code:: pycon
-
-          >>> from returns.context import RequiresContextResult
-          >>> from returns.result import Success, Failure
-
-          >>> assert RequiresContextResult(
-          ...    lambda _: Failure(1),
-          ... ).failure()(RequiresContextResult.empty) == 1
-
-          >>> RequiresContextResult(
-          ...    lambda _: Success(1),
-          ... ).failure()(RequiresContextResult.empty)
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
-
-        """
-        return lambda deps: self(deps).failure()
+        return RequiresContextResult(Success)
 
     @classmethod
     def from_result(
-        cls, inner_value: Result[_ValueType, _ErrorType],
-    ) -> 'RequiresContextResult[NoDeps, _ValueType, _ErrorType]':
+        cls, inner_value: Result[_NewValueType, _NewErrorType],
+    ) -> RequiresContextResult[_NewValueType, _NewErrorType, NoDeps]:
         """
         Creates new container with ``Result`` as a unit value.
 
@@ -473,7 +466,7 @@ class RequiresContextResult(
 
           >>> from returns.context import RequiresContextResult
           >>> from returns.result import Success, Failure
-          >>> deps = RequiresContextResult.empty
+          >>> deps = RequiresContextResult.no_args
 
           >>> assert RequiresContextResult.from_result(
           ...    Success(1),
@@ -490,8 +483,8 @@ class RequiresContextResult(
     def from_typecast(
         cls,
         inner_value:
-            'RequiresContext[_EnvType, Result[_NewValueType, _NewErrorType]]',
-    ) -> 'RequiresContextResult[_EnvType, _NewValueType, _NewErrorType]':
+            'RequiresContext[Result[_NewValueType, _NewErrorType], _EnvType]',
+    ) -> RequiresContextResult[_NewValueType, _NewErrorType, _EnvType]:
         """
         You might end up with ``RequiresContext[Result[...]]`` as a value.
 
@@ -507,19 +500,19 @@ class RequiresContextResult(
 
           >>> assert RequiresContextResult.from_typecast(
           ...     RequiresContext.from_value(Success(1)),
-          ... )(RequiresContextResult.empty) == Success(1)
+          ... )(RequiresContextResult.no_args) == Success(1)
 
           >>> assert RequiresContextResult.from_typecast(
           ...     RequiresContext.from_value(Failure(1)),
-          ... )(RequiresContextResult.empty) == Failure(1)
+          ... )(RequiresContextResult.no_args) == Failure(1)
 
         """
         return RequiresContextResult(inner_value)
 
     @classmethod
     def from_context(
-        cls, inner_value: 'RequiresContext[_EnvType, _FirstType]',
-    ) -> 'RequiresContextResult[_EnvType, _FirstType, Any]':
+        cls, inner_value: 'RequiresContext[_NewValueType, _NewEnvType]',
+    ) -> RequiresContextResult[_NewValueType, Any, _NewEnvType]:
         """
         Creates new container from ``RequiresContext`` as a success unit.
 
@@ -536,8 +529,8 @@ class RequiresContextResult(
 
     @classmethod
     def from_failed_context(
-        cls, inner_value: 'RequiresContext[_EnvType, _FirstType]',
-    ) -> 'RequiresContextResult[_EnvType, Any, _FirstType]':
+        cls, inner_value: 'RequiresContext[_NewValueType, _NewEnvType]',
+    ) -> RequiresContextResult[Any, _NewValueType, _NewEnvType]:
         """
         Creates new container from ``RequiresContext`` as a failure unit.
 
@@ -552,10 +545,35 @@ class RequiresContextResult(
         """
         return RequiresContextResult(lambda deps: Failure(inner_value(deps)))
 
+    @classmethod
+    def from_result_context(
+        cls,
+        inner_value:
+            RequiresContextResult[_NewValueType, _NewErrorType, _NewEnvType],
+    ) -> RequiresContextResult[_NewValueType, _NewErrorType, _NewEnvType]:
+        """
+        Creates ``RequiresContextResult`` from another instance of it.
+
+        .. code:: python
+
+          >>> from returns.context import ReaderResult
+          >>> from returns.result import Success, Failure
+
+          >>> assert ReaderResult.from_result_context(
+          ...     ReaderResult.from_value(1),
+          ... )(...) == Success(1)
+
+          >>> assert ReaderResult.from_result_context(
+          ...     ReaderResult.from_failure(1),
+          ... )(...) == Failure(1)
+
+        """
+        return inner_value
+
     @classmethod
     def from_value(
         cls, inner_value: _FirstType,
-    ) -> 'RequiresContextResult[NoDeps, _FirstType, Any]':
+    ) -> RequiresContextResult[_FirstType, Any, NoDeps]:
         """
         Creates new container with ``Success(inner_value)`` as a unit value.
 
@@ -571,7 +589,7 @@ class RequiresContextResult(
     @classmethod
     def from_failure(
         cls, inner_value: _FirstType,
-    ) -> 'RequiresContextResult[NoDeps, Any, _FirstType]':
+    ) -> RequiresContextResult[Any, _FirstType, NoDeps]:
         """
         Creates new container with ``Failure(inner_value)`` as a unit value.
 
@@ -584,84 +602,16 @@ class RequiresContextResult(
         """
         return RequiresContextResult(lambda _: Failure(inner_value))
 
-    @classmethod
-    def from_iterable(
-        cls,
-        inner_value:
-            Iterable['RequiresContextResult[_EnvType, _ValueType, _ErrorType]'],
-    ) -> 'RequiresContextResult[_EnvType, Sequence[_ValueType], _ErrorType]':
-        """
-        Transforms an iterable of ``RequiresContextResult`` containers.
-
-        Returns a single container with multiple elements inside.
-
-        .. code:: python
-
-          >>> from returns.context import RequiresContextResult
-          >>> from returns.result import Success, Failure
-
-          >>> assert RequiresContextResult.from_iterable([
-          ...    RequiresContextResult.from_value(1),
-          ...    RequiresContextResult.from_value(2),
-          ... ])(...) == Success((1, 2))
-
-          >>> assert RequiresContextResult.from_iterable([
-          ...    RequiresContextResult.from_value(1),
-          ...    RequiresContextResult.from_failure('a'),
-          ... ])(...) == Failure('a')
-
-          >>> assert RequiresContextResult.from_iterable([
-          ...    RequiresContextResult.from_failure('a'),
-          ...    RequiresContextResult.from_value(1),
-          ... ])(...) == Failure('a')
-
-        """
-        return iterable(cls, inner_value)
-
-
-@final
-class ContextResult(Immutable, Generic[_EnvType], metaclass=ABCMeta):
-    """
-    Helpers that can be used to work with ``RequiresContextResult`` container.
-
-    Related to :class:`returns.context.Context`, refer there for the docs.
-    """
-
-    __slots__ = ()
-
-    @classmethod
-    def ask(cls) -> RequiresContextResult[_EnvType, _EnvType, Any]:
-        """
-        Is used to get the current dependencies inside the call stack.
-
-        Similar to :meth:`returns.context.Context.ask`,
-        but returns ``Result`` instead of a regular value.
-
-        Please, refer to the docs there to learn how to use it.
-
-        One important note that is worth duplicating here:
-        you might need to provide ``_EnvType`` explicitly,
-        so ``mypy`` will know about it statically.
-
-        .. code:: python
-
-          >>> from returns.context import ContextResult
-          >>> from returns.result import Success
-          >>> assert ContextResult[int].ask().map(str)(1) == Success('1')
-
-        """
-        return RequiresContextResult(Success)
-
 
 # Aliases:
 
 #: Alias for a popular case when ``Result`` has ``Exception`` as error type.
 RequiresContextResultE = RequiresContextResult[
-    _EnvType, _ValueType, Exception,
+    _ValueType, Exception, _EnvType,
 ]
 
 #: Alias to save you some typing. Uses original name from Haskell.
-ReaderResult = RequiresContextResult[_EnvType, _ValueType, _ErrorType]
+ReaderResult = RequiresContextResult
 
 #: Alias to save you some typing. Has ``Exception`` as error type.
-ReaderResultE = RequiresContextResult[_EnvType, _ValueType, Exception]
+ReaderResultE = RequiresContextResult[_ValueType, Exception, _EnvType]
