@@ -1,25 +1,21 @@
 from functools import wraps
-from typing import (
-    Any,
-    Awaitable,
-    Callable,
-    ClassVar,
-    Coroutine,
-    Generator,
-    Generic,
-    Iterable,
-    Sequence,
-    Type,
-    TypeVar,
-    Union,
-)
+from typing import Any, Awaitable, Callable, Coroutine, Generator, TypeVar
 
 from typing_extensions import final
 
-from returns._generated.futures import _future, _future_result
-from returns._generated.iterable import iterable
+from returns._internal.futures import _future, _future_result
+from returns.interfaces.specific.future import FutureBased1
+from returns.interfaces.specific.future_result import FutureResultBased2
 from returns.io import IO, IOResult
 from returns.primitives.container import BaseContainer
+from returns.primitives.hkt import (
+    Kind1,
+    Kind2,
+    SupportsKind1,
+    SupportsKind2,
+    dekind,
+)
+from returns.primitives.reawaitable import ReAwaitable
 from returns.result import Failure, Result, Success
 
 # Definitions:
@@ -56,7 +52,11 @@ async def async_identity(instance: _FirstType) -> _FirstType:
 # ======
 
 @final
-class Future(BaseContainer, Generic[_ValueType]):
+class Future(
+    BaseContainer,
+    SupportsKind1['Future', _ValueType],
+    FutureBased1[_ValueType],
+):
     """
     Container to easily compose ``async`` functions.
 
@@ -74,16 +74,14 @@ class Future(BaseContainer, Generic[_ValueType]):
 
     Is not related to ``asyncio.Future`` in any kind.
 
-    Tradeoffs
-    ---------
+    .. rubric:: Tradeoffs
 
     Due to possible performance issues we move all coroutines definitions
     to a separate module.
 
     See also:
-        https://gcanti.github.io/fp-ts/modules/Task.ts.html
-        https://zio.dev/docs/overview/overview_basic_concurrency
-        https://github.com/correl/typesafe-monads/blob/master/monads/future.py
+        - https://gcanti.github.io/fp-ts/modules/Task.ts.html
+        - https://zio.dev/docs/overview/overview_basic_concurrency
 
     """
 
@@ -106,9 +104,9 @@ class Future(BaseContainer, Generic[_ValueType]):
           >>> assert anyio.run(container.awaitable) == IO(2)
 
         """
-        super().__init__(inner_value)
+        super().__init__(ReAwaitable(inner_value))
 
-    def __await__(self) -> Generator[Any, Any, IO[_ValueType]]:
+    def __await__(self) -> Generator[None, None, IO[_ValueType]]:
         """
         By defining this magic method we make ``Future`` awaitable.
 
@@ -130,8 +128,8 @@ class Future(BaseContainer, Generic[_ValueType]):
         to indicate that the computation was impure.
 
         See also:
-            https://docs.python.org/3/library/asyncio-task.html#awaitables
-            https://www.python.org/dev/peps/pep-0492/#new-abstract-base-classes
+            - https://docs.python.org/3/library/asyncio-task.html#awaitables
+            - https://bit.ly/2SfayNc
 
         """
         return self.awaitable().__await__()  # noqa: WPS609
@@ -186,7 +184,7 @@ class Future(BaseContainer, Generic[_ValueType]):
 
     def apply(
         self,
-        container: 'Future[Callable[[_ValueType], _NewValueType]]',
+        container: Kind1['Future', Callable[[_ValueType], _NewValueType]],
     ) -> 'Future[_NewValueType]':
         """
         Calls a wrapped function in a container on this container.
@@ -206,11 +204,11 @@ class Future(BaseContainer, Generic[_ValueType]):
           ... ) == IO('1b')
 
         """
-        return Future(_future.async_apply(container, self._inner_value))
+        return Future(_future.async_apply(dekind(container), self._inner_value))
 
     def bind(
         self,
-        function: Callable[[_ValueType], 'Future[_NewValueType]'],
+        function: Callable[[_ValueType], Kind1['Future', _NewValueType]],
     ) -> 'Future[_NewValueType]':
         """
         Applies 'function' to the result of a previous calculation.
@@ -234,9 +232,15 @@ class Future(BaseContainer, Generic[_ValueType]):
         """
         return Future(_future.async_bind(function, self._inner_value))
 
+    #: Alias for `bind` method. Part of the `FutureBasedN` interface.
+    bind_future = bind
+
     def bind_async(
         self,
-        function: Callable[[_ValueType], Awaitable['Future[_NewValueType]']],
+        function: Callable[
+            [_ValueType],
+            Awaitable[Kind1['Future', _NewValueType]],
+        ],
     ) -> 'Future[_NewValueType]':
         """
         Compose a container and ``async`` function returning a container.
@@ -261,6 +265,9 @@ class Future(BaseContainer, Generic[_ValueType]):
         """
         return Future(_future.async_bind_async(function, self._inner_value))
 
+    #: Alias for `bind_async` method. Part of the `FutureBasedN` interface.
+    bind_async_future = bind_async
+
     def bind_awaitable(
         self,
         function: Callable[[_ValueType], 'Awaitable[_NewValueType]'],
@@ -339,12 +346,11 @@ class Future(BaseContainer, Generic[_ValueType]):
         return Future(async_identity(inner_value))
 
     @classmethod
-    def from_iterable(
-        cls,
-        inner_value: Iterable['Future[_ValueType]'],
-    ) -> 'Future[Sequence[_ValueType]]':
+    def from_future(
+        cls, inner_value: 'Future[_NewValueType]',
+    ) -> 'Future[_NewValueType]':
         """
-        Transforms an iterable of ``Future`` containers into a single container.
+        Creates a new ``Future`` from the existing one.
 
         .. code:: python
 
@@ -352,13 +358,12 @@ class Future(BaseContainer, Generic[_ValueType]):
           >>> from returns.future import Future
           >>> from returns.io import IO
 
-          >>> assert anyio.run(Future.from_iterable([
-          ...    Future.from_value(1),
-          ...    Future.from_value(2),
-          ... ]).awaitable) == IO((1, 2))
+          >>> future = Future.from_value(1)
+          >>> assert anyio.run(Future.from_future(future).awaitable) == IO(1)
 
+        Part of the ``FutureBasedN`` interface.
         """
-        return iterable(cls, inner_value)
+        return inner_value
 
     @classmethod
     def from_io(cls, inner_value: IO[_NewValueType]) -> 'Future[_NewValueType]':
@@ -382,8 +387,8 @@ class Future(BaseContainer, Generic[_ValueType]):
     @classmethod
     def from_future_result(
         cls,
-        inner_value: 'FutureResult[_ValueType, _ErrorType]',
-    ) -> 'Future[Result[_ValueType, _ErrorType]]':
+        inner_value: 'FutureResult[_NewValueType, _NewErrorType]',
+    ) -> 'Future[Result[_NewValueType, _NewErrorType]]':
         """
         Creates ``Future[Result[a, b]]`` instance from ``FutureResult[a, b]``.
 
@@ -423,7 +428,7 @@ def future(
       >>> @future
       ... async def test(x: int) -> int:
       ...     return x + 1
-      ...
+
       >>> assert anyio.run(test(1).awaitable) == IO(2)
 
     Requires our :ref:`mypy plugin <mypy-plugins>`.
@@ -482,7 +487,11 @@ def asyncify(function: Callable[..., _ValueType]) -> Callable[
 # ============
 
 @final
-class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
+class FutureResult(
+    BaseContainer,
+    SupportsKind2['FutureResult', _ValueType, _ErrorType],
+    FutureResultBased2[_ValueType, _ErrorType],
+):
     """
     Container to easily compose ``async`` functions.
 
@@ -502,20 +511,17 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
     ``FutureResult`` has a lot of composition helpers
     to turn complex nested operations into a one function calls.
 
-    Tradeoffs
-    ~~~~~~~~~
+    .. rubric:: Tradeoffs
 
     Due to possible performance issues we move all coroutines definitions
     to a separate module.
 
     See also:
-        https://gcanti.github.io/fp-ts/modules/TaskEither.ts.html
-        https://zio.dev/docs/overview/overview_basic_concurrency
+        - https://gcanti.github.io/fp-ts/modules/TaskEither.ts.html
+        - https://zio.dev/docs/overview/overview_basic_concurrency
 
     """
 
-    outer: ClassVar[Type[Future]] = Future
-
     _inner_value: Awaitable[Result[_ValueType, _ErrorType]]
 
     def __init__(
@@ -539,10 +545,10 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
           >>> assert anyio.run(container.awaitable) == IOSuccess(2)
 
         """
-        super().__init__(inner_value)
+        super().__init__(ReAwaitable(inner_value))
 
     def __await__(self) -> Generator[
-        Any, Any, IOResult[_ValueType, _ErrorType],
+        None, None, IOResult[_ValueType, _ErrorType],
     ]:
         """
         By defining this magic method we make ``FutureResult`` awaitable.
@@ -565,8 +571,8 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
         to indicate that the computation was impure and can fail.
 
         See also:
-            https://docs.python.org/3/library/asyncio-task.html#awaitables
-            https://www.python.org/dev/peps/pep-0492/#new-abstract-base-classes
+            - https://docs.python.org/3/library/asyncio-task.html#awaitables
+            - https://bit.ly/2SfayNc
 
         """
         return self.awaitable().__await__()  # noqa: WPS609
@@ -593,6 +599,28 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
         """
         return IOResult.from_result(await self._inner_value)
 
+    def swap(self) -> 'FutureResult[_ErrorType, _ValueType]':
+        """
+        Swaps value and error types.
+
+        So, values become errors and errors become values.
+        It is useful when you have to work with errors a lot.
+        And since we have a lot of ``.bind_`` related methods
+        and only a single ``.lash``.
+        It is easier to work with values than with errors.
+
+        .. code:: python
+
+          >>> import anyio
+          >>> from returns.future import FutureSuccess, FutureFailure
+          >>> from returns.io import IOSuccess, IOFailure
+
+          >>> assert anyio.run(FutureSuccess(1).swap) == IOFailure(1)
+          >>> assert anyio.run(FutureFailure(1).swap) == IOSuccess(1)
+
+        """
+        return FutureResult(_future_result.async_swap(self._inner_value))
+
     def map(  # noqa: WPS125
         self,
         function: Callable[[_ValueType], _NewValueType],
@@ -628,8 +656,11 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
 
     def apply(
         self,
-        container:
-            'FutureResult[Callable[[_ValueType], _NewValueType], _ErrorType]',
+        container: Kind2[
+            'FutureResult',
+            Callable[[_ValueType], _NewValueType],
+            _ErrorType,
+        ],
     ) -> 'FutureResult[_NewValueType, _ErrorType]':
         """
         Calls a wrapped function in a container on this container.
@@ -654,22 +685,27 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
           ...     ).awaitable,
           ... ) == IOFailure(1)
 
-          >>> assert isinstance(anyio.run(
+          >>> assert anyio.run(
           ...     FutureResult.from_value(1).apply(
-          ...         FutureResult.from_failure(appliable),
+          ...         FutureResult.from_failure(2),
           ...     ).awaitable,
-          ... ), IOResult.failure_type)
+          ... ) == IOFailure(2)
+          >>> assert anyio.run(
+          ...     FutureResult.from_failure(1).apply(
+          ...         FutureResult.from_failure(2),
+          ...     ).awaitable,
+          ... ) == IOFailure(1)
 
         """
         return FutureResult(_future_result.async_apply(
-            container, self._inner_value,
+            dekind(container), self._inner_value,
         ))
 
     def bind(
         self,
         function: Callable[
             [_ValueType],
-            'FutureResult[_NewValueType, _ErrorType]',
+            Kind2['FutureResult', _NewValueType, _ErrorType],
         ],
     ) -> 'FutureResult[_NewValueType, _ErrorType]':
         """
@@ -699,11 +735,15 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
             function, self._inner_value,
         ))
 
+    #: Alias for `bind` method.
+    #: Part of the `FutureResultBasedN` interface.
+    bind_future_result = bind
+
     def bind_async(
         self,
         function: Callable[
             [_ValueType],
-            Awaitable['FutureResult[_NewValueType, _ErrorType]'],
+            Awaitable[Kind2['FutureResult', _NewValueType, _ErrorType]],
         ],
     ) -> 'FutureResult[_NewValueType, _ErrorType]':
         """
@@ -734,6 +774,10 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
             function, self._inner_value,
         ))
 
+    #: Alias for `bind_async` method.
+    #: Part of the `FutureResultBasedN` interface.
+    bind_async_future_result = bind_async
+
     def bind_awaitable(
         self,
         function: Callable[[_ValueType], Awaitable[_NewValueType]],
@@ -912,69 +956,6 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
             function, self._inner_value,
         ))
 
-    def unify(
-        self,
-        function: Callable[
-            [_ValueType], 'FutureResult[_NewValueType, _NewErrorType]',
-        ],
-    ) -> 'FutureResult[_NewValueType, Union[_ErrorType, _NewErrorType]]':
-        """
-        Composes successful container with a function that returns a container.
-
-        Similar to :meth:`~FutureResult.bind` but has different type.
-        It returns ``FutureResult[ValueType, Union[ErrorType, NewErrorType]]``
-        instead of ``FutureResult[ValueType, ErrorType]``.
-
-        So, it can be more useful in some situations.
-        Probably with specific exceptions.
-
-        .. code:: python
-
-          >>> import anyio
-          >>> from returns.future import FutureResult
-          >>> from returns.io import IOSuccess, IOFailure
-
-          >>> def bindable(x: int) -> FutureResult[int, str]:
-          ...    return FutureResult.from_value(x + 1)
-
-          >>> assert anyio.run(
-          ...     FutureResult.from_value(1).unify(bindable).awaitable,
-          ... ) == IOSuccess(2)
-          >>> assert anyio.run(
-          ...     FutureResult.from_failure(1).unify(bindable).awaitable,
-          ... ) == IOFailure(1)
-
-        """
-        return self.bind(function)  # type: ignore
-
-    def fix(
-        self,
-        function: Callable[[_ErrorType], _NewValueType],
-    ) -> 'FutureResult[_NewValueType, _ErrorType]':
-        """
-        Composes failed container with a pure function to fix the failure.
-
-        .. code:: python
-
-          >>> import anyio
-          >>> from returns.future import FutureResult
-          >>> from returns.io import IOSuccess
-
-          >>> def fixable(arg: int) -> int:
-          ...      return arg + 1
-
-          >>> assert anyio.run(
-          ...     FutureResult.from_value(1).fix(fixable).awaitable,
-          ... ) == IOSuccess(1)
-          >>> assert anyio.run(
-          ...     FutureResult.from_failure(1).fix(fixable).awaitable,
-          ... ) == IOSuccess(2)
-
-        """
-        return FutureResult(_future_result.async_fix(
-            function, self._inner_value,
-        ))
-
     def alt(
         self,
         function: Callable[[_ErrorType], _NewErrorType],
@@ -988,14 +969,14 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
           >>> from returns.future import FutureResult
           >>> from returns.io import IOSuccess, IOFailure
 
-          >>> def fixable(arg: int) -> int:
+          >>> def altable(arg: int) -> int:
           ...      return arg + 1
 
           >>> assert anyio.run(
-          ...     FutureResult.from_value(1).alt(fixable).awaitable,
+          ...     FutureResult.from_value(1).alt(altable).awaitable,
           ... ) == IOSuccess(1)
           >>> assert anyio.run(
-          ...     FutureResult.from_failure(1).alt(fixable).awaitable,
+          ...     FutureResult.from_failure(1).alt(altable).awaitable,
           ... ) == IOFailure(2)
 
         """
@@ -1003,11 +984,11 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
             function, self._inner_value,
         ))
 
-    def rescue(
+    def lash(
         self,
         function: Callable[
             [_ErrorType],
-            'FutureResult[_ValueType, _NewErrorType]',
+            Kind2['FutureResult', _ValueType, _NewErrorType],
         ],
     ) -> 'FutureResult[_ValueType, _NewErrorType]':
         """
@@ -1019,81 +1000,57 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
           >>> from returns.future import FutureResult
           >>> from returns.io import IOSuccess
 
-          >>> def rescuable(x: int) -> FutureResult[int, str]:
+          >>> def lashable(x: int) -> FutureResult[int, str]:
           ...    return FutureResult.from_value(x + 1)
 
           >>> assert anyio.run(
-          ...     FutureResult.from_value(1).rescue(rescuable).awaitable,
+          ...     FutureResult.from_value(1).lash(lashable).awaitable,
           ... ) == IOSuccess(1)
           >>> assert anyio.run(
-          ...     FutureResult.from_failure(1).rescue(rescuable).awaitable,
+          ...     FutureResult.from_failure(1).lash(lashable).awaitable,
           ... ) == IOSuccess(2)
 
         """
-        return FutureResult(_future_result.async_rescue(
+        return FutureResult(_future_result.async_lash(
             function, self._inner_value,
         ))
 
-    def value_or(
+    def compose_result(
         self,
-        default_value: _NewValueType,
-    ) -> Awaitable[IO[Union[_ValueType, _NewValueType]]]:
+        function: Callable[
+            [Result[_ValueType, _ErrorType]],
+            Kind2['FutureResult', _NewValueType, _ErrorType],
+        ],
+    ) -> 'FutureResult[_NewValueType, _ErrorType]':
         """
-        Get value or default value.
+        Composes inner ``Result`` with ``FutureResult`` returning function.
+
+        Can be useful when you need an access to both states of the result.
 
         .. code:: python
 
           >>> import anyio
           >>> from returns.future import FutureResult
-          >>> from returns.io import IO
+          >>> from returns.io import IOSuccess, IOFailure
+          >>> from returns.result import Result
 
-          >>> async def main():
-          ...     first = await FutureResult.from_value(1).value_or(2)
-          ...     second = await FutureResult.from_failure(3).value_or(4)
-          ...     return first, second
+          >>> def count(container: Result[int, int]) -> FutureResult[int, int]:
+          ...     return FutureResult.from_result(
+          ...         container.map(lambda x: x + 1).alt(abs),
+          ...     )
 
-          >>> assert anyio.run(main) == (IO(1), IO(4))
+          >>> assert anyio.run(
+          ...    FutureResult.from_value(1).compose_result, count,
+          ... ) == IOSuccess(2)
+          >>> assert anyio.run(
+          ...    FutureResult.from_failure(-1).compose_result, count,
+          ... ) == IOFailure(1)
 
         """
-        return _future_result.async_value_or(self, default_value)
-
-    def unwrap(self) -> Awaitable[IO[_ValueType]]:
-        """
-        Get value or raise exception.
-
-        .. code:: pycon
-
-          >>> import anyio
-          >>> from returns.future import FutureResult
-          >>> from returns.io import IO
-          >>> assert anyio.run(FutureResult.from_value(1).unwrap) == IO(1)
-
-          >>> anyio.run(FutureResult.from_failure(1).unwrap)
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
-
-        """
-        return _future_result.async_unwrap(self)
-
-    def failure(self) -> Awaitable[IO[_ErrorType]]:
-        """
-        Get failed value or raise exception.
-
-        .. code:: pycon
-
-          >>> import anyio
-          >>> from returns.future import FutureResult
-          >>> from returns.io import IO
-          >>> assert anyio.run(FutureResult.from_failure(1).failure) == IO(1)
-
-          >>> anyio.run(FutureResult.from_value(1).failure)
-          Traceback (most recent call last):
-            ...
-          returns.primitives.exceptions.UnwrapFailedError
-
-        """
-        return _future_result.async_failure(self)
+        return FutureResult(_future_result.async_compose_result(
+            function,
+            self._inner_value,
+        ))
 
     @classmethod
     def from_typecast(
@@ -1171,6 +1128,31 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
         """
         return FutureResult(_future_result.async_from_failure(inner_value))
 
+    @classmethod
+    def from_future_result(
+        cls,
+        inner_value: 'FutureResult[_NewValueType, _NewErrorType]',
+    ) -> 'FutureResult[_NewValueType, _NewErrorType]':
+        """
+        Creates new ``FutureResult`` from existing one.
+
+        .. code:: python
+
+          >>> import anyio
+          >>> from returns.io import IOSuccess
+          >>> from returns.future import FutureResult
+
+          >>> async def main():
+          ...     assert await FutureResult.from_future_result(
+          ...         FutureResult.from_value(1),
+          ...     ) == IOSuccess(1)
+
+          >>> anyio.run(main)
+
+        Part of the ``FutureResultLikeN`` interface.
+        """
+        return inner_value
+
     @classmethod
     def from_io(
         cls,
@@ -1322,42 +1304,47 @@ class FutureResult(BaseContainer, Generic[_ValueType, _ErrorType]):
         """
         return FutureResult(async_identity(Failure(inner_value)))
 
-    @classmethod
-    def from_iterable(
-        cls,
-        inner_value: Iterable['FutureResult[_ValueType, _ErrorType]'],
-    ) -> 'FutureResult[Sequence[_ValueType], _ErrorType]':
-        """
-        Transforms an iterable of ``FutureResult`` containers.
 
-        Returns a single container with multiple elements inside.
+def FutureSuccess(  # noqa: N802
+    inner_value: _NewValueType,
+) -> FutureResult[_NewValueType, Any]:
+    """
+    Public unit function to create successful ``FutureResult`` objects.
 
-        .. code:: python
+    Is the same as :meth:`~FutureResult.from_value`.
 
-          >>> import anyio
-          >>> from returns.future import FutureResult
-          >>> from returns.io import IOSuccess, IOFailure
+    .. code:: python
 
-          >>> all_success = FutureResult.from_iterable([
-          ...    FutureResult.from_value(1),
-          ...    FutureResult.from_value(2),
-          ... ])
-          >>> assert anyio.run(all_success.awaitable) == IOSuccess((1, 2))
+      >>> import anyio
+      >>> from returns.future import FutureResult, FutureSuccess
 
-          >>> mixed = FutureResult.from_iterable([
-          ...     FutureResult.from_value(1),
-          ...     FutureResult.from_failure('a'),
-          ... ])
-          >>> assert anyio.run(mixed.awaitable) == IOFailure('a')
+      >>> assert anyio.run(FutureSuccess(1).awaitable) == anyio.run(
+      ...     FutureResult.from_value(1).awaitable,
+      ... )
 
-          >>> mixed = FutureResult.from_iterable([
-          ...     FutureResult.from_failure('a'),
-          ...     FutureResult.from_value(1),
-          ... ])
-          >>> assert anyio.run(mixed.awaitable) == IOFailure('a')
+    """
+    return FutureResult.from_value(inner_value)
 
-        """
-        return iterable(cls, inner_value)
+
+def FutureFailure(  # noqa: N802
+    inner_value: _NewErrorType,
+) -> FutureResult[Any, _NewErrorType]:
+    """
+    Public unit function to create failed ``FutureResult`` objects.
+
+    Is the same as :meth:`~FutureResult.from_failure`.
+
+    .. code:: python
+
+      >>> import anyio
+      >>> from returns.future import FutureResult, FutureFailure
+
+      >>> assert anyio.run(FutureFailure(1).awaitable) == anyio.run(
+      ...     FutureResult.from_failure(1).awaitable,
+      ... )
+
+    """
+    return FutureResult.from_failure(inner_value)
 
 
 # Aliases:
