@@ -1,21 +1,19 @@
 from abc import ABCMeta
 from functools import wraps
-from typing import (
-    Any,
-    Callable,
-    ClassVar,
-    Generic,
-    Iterable,
-    Sequence,
-    Type,
-    TypeVar,
-    Union,
-)
+from inspect import FrameInfo
+from typing import Any, Callable, ClassVar, List, Optional, Type, TypeVar, Union
 
 from typing_extensions import final
 
-from returns._generated.iterable import iterable
-from returns.primitives.container import BaseContainer
+from returns.interfaces.specific import io, ioresult
+from returns.primitives.container import BaseContainer, container_equality
+from returns.primitives.hkt import (
+    Kind1,
+    Kind2,
+    SupportsKind1,
+    SupportsKind2,
+    dekind,
+)
 from returns.result import Failure, Result, Success
 
 _ValueType = TypeVar('_ValueType', covariant=True)
@@ -30,7 +28,11 @@ _FirstType = TypeVar('_FirstType')
 _SecondType = TypeVar('_SecondType')
 
 
-class IO(BaseContainer, Generic[_ValueType]):
+class IO(
+    BaseContainer,
+    SupportsKind1['IO', _ValueType],
+    io.IOLike1[_ValueType],
+):
     """
     Explicit container for impure function results.
 
@@ -51,13 +53,16 @@ class IO(BaseContainer, Generic[_ValueType]):
     Like DB access or network operations.
 
     See also:
-        https://dev.to/gcanti/getting-started-with-fp-ts-io-36p6
-        https://gist.github.com/chris-taylor/4745921
+        - https://dev.to/gcanti/getting-started-with-fp-ts-io-36p6
+        - https://gist.github.com/chris-taylor/4745921
 
     """
 
     _inner_value: _ValueType
 
+    #: Typesafe equality comparison with other `Result` objects.
+    equals = container_equality
+
     def __init__(self, inner_value: _ValueType) -> None:
         """
         Public constructor for this type. Also required for typing.
@@ -65,8 +70,7 @@ class IO(BaseContainer, Generic[_ValueType]):
         .. code:: python
 
           >>> from returns.io import IO
-          >>> str(IO(1))
-          '<IO: 1>'
+          >>> assert str(IO(1)) == '<IO: 1>'
 
         """
         super().__init__(inner_value)
@@ -95,7 +99,7 @@ class IO(BaseContainer, Generic[_ValueType]):
 
     def apply(
         self,
-        container: 'IO[Callable[[_ValueType], _NewValueType]]',
+        container: Kind1['IO', Callable[[_ValueType], _NewValueType]],
     ) -> 'IO[_NewValueType]':
         """
         Calls a wrapped function in a container on this container.
@@ -119,10 +123,11 @@ class IO(BaseContainer, Generic[_ValueType]):
           >>> assert IO('b').apply(IO('a').apply(IO(appliable))) == IO('ab')
 
         """
-        return self.map(container._inner_value)  # noqa: WPS437
+        return self.map(dekind(container)._inner_value)  # noqa: WPS437
 
     def bind(
-        self, function: Callable[[_ValueType], 'IO[_NewValueType]'],
+        self,
+        function: Callable[[_ValueType], Kind1['IO', _NewValueType]],
     ) -> 'IO[_NewValueType]':
         """
         Applies 'function' to the result of a previous calculation.
@@ -138,7 +143,10 @@ class IO(BaseContainer, Generic[_ValueType]):
           >>> assert IO('a').bind(bindable) == IO('ab')
 
         """
-        return function(self._inner_value)
+        return dekind(function(self._inner_value))
+
+    #: Alias for `bind` method. Part of the `IOLikeN` interface.
+    bind_io = bind
 
     @classmethod
     def from_value(cls, inner_value: _NewValueType) -> 'IO[_NewValueType]':
@@ -152,36 +160,31 @@ class IO(BaseContainer, Generic[_ValueType]):
           >>> from returns.io import IO
           >>> assert IO(1) == IO.from_value(1)
 
-        Part of the :class:`returns.primitives.interfaces.Applicative`
-        protocol.
+        Part of the :class:`returns.interfaces.applicative.ApplicativeN`
+        interface.
         """
         return IO(inner_value)
 
     @classmethod
-    def from_iterable(
-        cls,
-        inner_value: Iterable['IO[_ValueType]'],
-    ) -> 'IO[Sequence[_ValueType]]':
+    def from_io(cls, inner_value: 'IO[_NewValueType]') -> 'IO[_NewValueType]':
         """
-        Transforms an iterable of ``IO`` containers into a single container.
+        Unit function to construct new ``IO`` values from existing ``IO``.
 
         .. code:: python
 
           >>> from returns.io import IO
+          >>> assert IO(1) == IO.from_io(IO(1))
 
-          >>> assert IO.from_iterable([
-          ...    IO(1),
-          ...    IO(2),
-          ... ]) == IO((1, 2))
+        Part of the :class:`returns.interfaces.specific.IO.IOLikeN` interface.
 
         """
-        return iterable(cls, inner_value)
+        return inner_value
 
     @classmethod
     def from_ioresult(
         cls,
-        inner_value: 'IOResult[_NewValueType, _ErrorType]',
-    ) -> 'IO[Result[_NewValueType, _ErrorType]]':
+        inner_value: 'IOResult[_NewValueType, _NewErrorType]',
+    ) -> 'IO[Result[_NewValueType, _NewErrorType]]':
         """
         Converts ``IOResult[a, b]`` back to ``IO[Result[a, b]]``.
 
@@ -234,14 +237,15 @@ def impure(
 
 class IOResult(
     BaseContainer,
-    Generic[_ValueType, _ErrorType],
+    SupportsKind2['IOResult', _ValueType, _ErrorType],
+    ioresult.IOResultBased2[_ValueType, _ErrorType],
     metaclass=ABCMeta,
 ):
     """
     Explicit container for impure function results that might fail.
 
-    Definition
-    ~~~~~~~~~~
+    .. rubric:: Definition
+
     This type is similar to :class:`returns.result.Result`.
     This basically a more useful version of ``IO[Result[a, b]]``.
     Use this type for ``IO`` computations that might fail.
@@ -281,11 +285,10 @@ class IOResult(
     See also:
         https://github.com/gcanti/fp-ts/blob/master/docs/modules/IOEither.ts.md
 
+    .. rubric:: Implementation
 
-    Implementation
-    ~~~~~~~~~~~~~~
     This class contains all the methods that can be delegated to ``Result``.
-    But, some methods have ``raise NotImplementedError`` which means
+    But, some methods are not implemented which means
     that we have to use special :class:`~_IOSuccess` and :class:`~_IOFailure`
     implementation details to correctly handle these callbacks.
 
@@ -293,8 +296,6 @@ class IOResult(
 
     """
 
-    outer: ClassVar[Type[IO]] = IO
-
     _inner_value: Result[_ValueType, _ErrorType]
 
     # These two are required for projects like `classes`:
@@ -303,6 +304,9 @@ class IOResult(
     #: Failure type that is used to represent the failed computation.
     failure_type: ClassVar[Type['_IOFailure']]
 
+    #: Typesafe equality comparison with other `IOResult` objects.
+    equals = container_equality
+
     def __init__(self, inner_value: Result[_ValueType, _ErrorType]) -> None:
         """
         Private type constructor.
@@ -312,6 +316,44 @@ class IOResult(
         """
         super().__init__(inner_value)
 
+    def __repr__(self) -> str:
+        """
+        Custom ``str`` representation for better readability.
+
+        .. code:: python
+
+          >>> from returns.io import IOSuccess, IOFailure
+          >>> assert str(IOSuccess(1)) == '<IOResult: <Success: 1>>'
+          >>> assert repr(IOSuccess(1)) == '<IOResult: <Success: 1>>'
+          >>> str(IOFailure(ValueError('wrong!')))
+          '<IOResult: <Failure: wrong!>>'
+
+        """
+        return '<IOResult: {0}>'.format(str(self._inner_value))
+
+    @property
+    def trace(self) -> Optional[List[FrameInfo]]:
+        """Returns a stack trace when :func:`~IOFailure` was called."""
+        return self._inner_value.trace
+
+    def swap(self) -> 'IOResult[_ErrorType, _ValueType]':
+        """
+        Swaps value and error types.
+
+        So, values become errors and errors become values.
+        It is useful when you have to work with errors a lot.
+        And since we have a lot of ``.bind_`` related methods
+        and only a single ``.lash`` - it is easier to work with values.
+
+        .. code:: python
+
+          >>> from returns.io import IOSuccess, IOFailure
+          >>> assert IOSuccess(1).swap() == IOFailure(1)
+          >>> assert IOFailure(1).swap() == IOSuccess(1)
+
+        """
+        return self.from_result(self._inner_value.swap())
+
     def map(  # noqa: WPS125
         self, function: Callable[[_ValueType], _NewValueType],
     ) -> 'IOResult[_NewValueType, _ErrorType]':
@@ -328,15 +370,18 @@ class IOResult(
 
     def apply(
         self,
-        container:
-            'IOResult[Callable[[_ValueType], _NewValueType], _ErrorType]',
+        container: Kind2[
+            'IOResult',
+            Callable[[_ValueType], _NewValueType],
+            _ErrorType,
+        ],
     ) -> 'IOResult[_NewValueType, _ErrorType]':
         """
         Calls a wrapped function in a container on this container.
 
         .. code:: python
 
-          >>> from returns.io import IOSuccess, IOFailure, IOResult
+          >>> from returns.io import IOSuccess, IOFailure
 
           >>> def appliable(first: str) -> str:
           ...      return first + 'b'
@@ -348,11 +393,12 @@ class IOResult(
           ...     IOSuccess(appliable),
           ... ) == IOFailure('a')
 
-          >>> assert isinstance(IOSuccess('a').apply(
-          ...     IOFailure(appliable),
-          ... ), IOResult.failure_type)
+          >>> assert IOSuccess('a').apply(IOFailure(1)) == IOFailure(1)
+          >>> assert IOFailure('a').apply(IOFailure('b')) == IOFailure('a')
 
         """
+        if isinstance(self, self.failure_type):
+            return self
         if isinstance(container, self.success_type):
             return self.from_result(
                 self._inner_value.map(
@@ -365,7 +411,7 @@ class IOResult(
         self,
         function: Callable[
             [_ValueType],
-            'IOResult[_NewValueType, _ErrorType]',
+            Kind2['IOResult', _NewValueType, _ErrorType],
         ],
     ) -> 'IOResult[_NewValueType, _ErrorType]':
         """
@@ -384,13 +430,15 @@ class IOResult(
           >>> assert IOFailure('a').bind(bindable) == IOFailure('a')
 
         """
-        raise NotImplementedError
+
+    #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.
+    bind_ioresult = bind
 
     def bind_result(
         self,
         function: Callable[
             [_ValueType],
-            'Result[_NewValueType, _ErrorType]',
+            Result[_NewValueType, _ErrorType],
         ],
     ) -> 'IOResult[_NewValueType, _ErrorType]':
         """
@@ -415,7 +463,6 @@ class IOResult(
           >>> assert IOFailure('a').bind_result(bindable) == IOFailure('a')
 
         """
-        raise NotImplementedError
 
     def bind_io(
         self,
@@ -438,49 +485,6 @@ class IOResult(
           >>> assert IOFailure('a').bind_io(bindable) == IOFailure('a')
 
         """
-        raise NotImplementedError
-
-    def unify(
-        self,
-        function: Callable[
-            [_ValueType],
-            'IOResult[_NewValueType, _NewErrorType]',
-        ],
-    ) -> 'IOResult[_NewValueType, Union[_ErrorType, _NewErrorType]]':
-        """
-        Composes successful container with a function that returns a container.
-
-        .. code:: python
-
-          >>> from returns.io import IOResult, IOFailure, IOSuccess
-          >>> def bindable(string: str) -> IOResult[str, str]:
-          ...      if len(string) > 1:
-          ...          return IOSuccess(string + 'b')
-          ...      return IOFailure(string + 'c')
-
-          >>> assert IOSuccess('aa').unify(bindable) == IOSuccess('aab')
-          >>> assert IOSuccess('a').unify(bindable) == IOFailure('ac')
-          >>> assert IOFailure('a').unify(bindable) == IOFailure('a')
-
-        """
-        return self.bind(function)  # type: ignore
-
-    def fix(
-        self,
-        function: Callable[[_ErrorType], _NewValueType],
-    ) -> 'IOResult[_NewValueType, _ErrorType]':
-        """
-        Composes failed container with a pure function to fix the failure.
-
-        .. code:: python
-
-          >>> from returns.io import IOFailure, IOSuccess
-          >>> assert IOFailure('a').fix(
-          ...     lambda char: char + 'b',
-          ... ) == IOSuccess('ab')
-
-        """
-        return self.from_result(self._inner_value.fix(function))
 
     def alt(
         self,
@@ -497,11 +501,11 @@ class IOResult(
         """
         return self.from_result(self._inner_value.alt(function))
 
-    def rescue(
+    def lash(
         self,
         function: Callable[
             [_ErrorType],
-            'IOResult[_ValueType, _NewErrorType]',
+            Kind2['IOResult', _ValueType, _NewErrorType],
         ],
     ) -> 'IOResult[_ValueType, _NewErrorType]':
         """
@@ -510,17 +514,16 @@ class IOResult(
         .. code:: python
 
           >>> from returns.io import IOFailure, IOSuccess, IOResult
-          >>> def rescuable(state: str) -> IOResult[int, str]:
+          >>> def lashable(state: str) -> IOResult[int, str]:
           ...     if len(state) > 1:
           ...         return IOSuccess(len(state))
           ...     return IOFailure('oops')
 
-          >>> assert IOFailure('a').rescue(rescuable) == IOFailure('oops')
-          >>> assert IOFailure('abc').rescue(rescuable) == IOSuccess(3)
-          >>> assert IOSuccess('a').rescue(rescuable) == IOSuccess('a')
+          >>> assert IOFailure('a').lash(lashable) == IOFailure('oops')
+          >>> assert IOFailure('abc').lash(lashable) == IOSuccess(3)
+          >>> assert IOSuccess('a').lash(lashable) == IOSuccess('a')
 
         """
-        raise NotImplementedError
 
     def value_or(
         self,
@@ -543,6 +546,7 @@ class IOResult(
         Get value from successful container or raise exception for failed one.
 
         .. code:: pycon
+          :force:
 
           >>> from returns.io import IO, IOFailure, IOSuccess
           >>> assert IOSuccess(1).unwrap() == IO(1)
@@ -552,7 +556,7 @@ class IOResult(
             ...
           returns.primitives.exceptions.UnwrapFailedError
 
-        """
+        """  # noqa: RST399
         return IO(self._inner_value.unwrap())
 
     def failure(self) -> IO[_ErrorType]:
@@ -560,6 +564,7 @@ class IOResult(
         Get failed value from failed container or raise exception from success.
 
         .. code:: pycon
+          :force:
 
           >>> from returns.io import IO, IOFailure, IOSuccess
           >>> assert IOFailure(1).failure() == IO(1)
@@ -569,9 +574,37 @@ class IOResult(
             ...
           returns.primitives.exceptions.UnwrapFailedError
 
-        """
+        """  # noqa: RST399
         return IO(self._inner_value.failure())
 
+    def compose_result(
+        self,
+        function: Callable[
+            [Result[_ValueType, _ErrorType]],
+            Kind2['IOResult', _NewValueType, _ErrorType],
+        ],
+    ) -> 'IOResult[_NewValueType, _ErrorType]':
+        """
+        Composes inner ``Result`` with ``IOResult`` returning function.
+
+        Can be useful when you need an access to both states of the result.
+
+        .. code:: python
+
+          >>> from returns.io import IOResult, IOSuccess, IOFailure
+          >>> from returns.result import Result
+
+          >>> def count(container: Result[int, int]) -> IOResult[int, int]:
+          ...     return IOResult.from_result(
+          ...         container.map(lambda x: x + 1).alt(abs),
+          ...     )
+
+          >>> assert IOSuccess(1).compose_result(count) == IOSuccess(2)
+          >>> assert IOFailure(-1).compose_result(count) == IOFailure(1)
+
+        """
+        return dekind(function(self._inner_value))
+
     @classmethod
     def from_typecast(
         cls, inner_value: IO[Result[_NewValueType, _NewErrorType]],
@@ -646,6 +679,23 @@ class IOResult(
             return _IOSuccess(inner_value)
         return _IOFailure(inner_value)
 
+    @classmethod
+    def from_ioresult(
+        cls, inner_value: 'IOResult[_NewValueType, _NewErrorType]',
+    ) -> 'IOResult[_NewValueType, _NewErrorType]':
+        """
+        Creates ``IOResult`` from existing ``IOResult`` value.
+
+        .. code:: python
+
+          >>> from returns.io import IOResult, IOSuccess, IOFailure
+
+          >>> assert IOResult.from_ioresult(IOSuccess(1)) == IOSuccess(1)
+          >>> assert IOResult.from_ioresult(IOFailure(2)) == IOFailure(2)
+
+        """
+        return inner_value
+
     @classmethod
     def from_value(
         cls, inner_value: _NewValueType,
@@ -653,7 +703,6 @@ class IOResult(
         """
         One more value to create success unit values.
 
-        This is a part of :class:`returns.primitives.interfaces.Unitable`.
         It is useful as a united way to create a new value from any container.
 
         .. code:: python
@@ -674,7 +723,6 @@ class IOResult(
         """
         One more value to create failure unit values.
 
-        This is a part of :class:`returns.primitives.interfaces.Unitable`.
         It is useful as a united way to create a new value from any container.
 
         .. code:: python
@@ -688,40 +736,6 @@ class IOResult(
         """
         return IOFailure(inner_value)
 
-    @classmethod
-    def from_iterable(
-        cls,
-        inner_value: Iterable['IOResult[_ValueType, _ErrorType]'],
-    ) -> 'IOResult[Sequence[_ValueType], _ErrorType]':
-        """
-        Transforms an iterable of ``IOResult`` containers into a single one.
-
-        .. code:: python
-
-          >>> from returns.io import IOResult, IOSuccess, IOFailure
-
-          >>> assert IOResult.from_iterable([
-          ...    IOSuccess(1),
-          ...    IOSuccess(2),
-          ... ]) == IOSuccess((1, 2))
-
-          >>> assert IOResult.from_iterable([
-          ...     IOSuccess(1),
-          ...     IOFailure('a'),
-          ... ]) == IOFailure('a')
-
-          >>> assert IOResult.from_iterable([
-          ...     IOFailure('a'),
-          ...     IOSuccess(1),
-          ... ]) == IOFailure('a')
-
-        """
-        return iterable(cls, inner_value)
-
-    def __str__(self) -> str:
-        """Custom ``str`` representation for better readability."""
-        return '<IOResult: {0}>'.format(str(self._inner_value))
-
 
 @final
 class _IOFailure(IOResult):
@@ -748,6 +762,9 @@ class _IOFailure(IOResult):
         """Does nothing for ``IOFailure``."""
         return self
 
+    #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.
+    bind_ioresult = bind
+
     def bind_result(self, function):
         """Does nothing for ``IOFailure``."""
         return self
@@ -756,7 +773,7 @@ class _IOFailure(IOResult):
         """Does nothing for ``IOFailure``."""
         return self
 
-    def rescue(self, function):
+    def lash(self, function):
         """Composes this container with a function returning ``IOResult``."""
         return function(self._inner_value.failure())
 
@@ -786,6 +803,9 @@ class _IOSuccess(IOResult):
         """Composes this container with a function returning ``IOResult``."""
         return function(self._inner_value.unwrap())
 
+    #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.
+    bind_ioresult = bind
+
     def bind_result(self, function):
         """Binds ``Result`` returning function to current container."""
         return self.from_result(function(self._inner_value.unwrap()))
@@ -794,7 +814,7 @@ class _IOSuccess(IOResult):
         """Binds ``IO`` returning function to current container."""
         return self.from_io(function(self._inner_value.unwrap()))
 
-    def rescue(self, function):
+    def lash(self, function):
         """Does nothing for ``IOSuccess``."""
         return self
 
@@ -810,8 +830,7 @@ def IOSuccess(  # noqa: N802
     .. code:: python
 
       >>> from returns.io import IOSuccess
-      >>> str(IOSuccess(1))
-      '<IOResult: <Success: 1>>'
+      >>> assert str(IOSuccess(1)) == '<IOResult: <Success: 1>>'
 
     """
     return _IOSuccess(Success(inner_value))
@@ -826,8 +845,7 @@ def IOFailure(  # noqa: N802
     .. code:: python
 
       >>> from returns.io import IOFailure
-      >>> str(IOFailure(1))
-      '<IOResult: <Failure: 1>>'
+      >>> assert str(IOFailure(1)) == '<IOResult: <Failure: 1>>'
 
     """
     return _IOFailure(Failure(inner_value))
@@ -874,7 +892,6 @@ def impure_safe(
     and :func:`returns.result.safe` decorators.
 
     Requires our :ref:`mypy plugin <mypy-plugins>`.
-
     """
     @wraps(function)
     def decorator(*args, **kwargs):
